# Warewulf shared library
#
# This script is a component of Warewulf,
# http://www.runlevelzero.net/greg/warewulf
#
#########################################################################
#
# Copyright (c) 2003, The Regents of the University of California, through
# Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The GNU GPL Document can be found at:
# http://www.gnu.org/copyleft/gpl.html
#
#########################################################################
#
# Written and maintained by:
#       Greg Kurtzer, <gmkurtzer@lbl.gov>

my $debug = ();
#$debug = 1;

package Warewulf;
use Exporter;
use Socket;
use IO::Socket;
use File::Path;
@ISA = ('Exporter');

@EXPORT = qw (
   &node_info
   &master_info
   &group_info
   &node_conf_sanity
   &gen_node_ips
   &node_status
   &gen_hosts 
   &write_hosts 
   &users_filter
   &client_conf
   &ipcalc_network
   &generate_random_string
   &node_addr
   &ipcalc_network
   &get_net_conf
   &pxe_update
);

###############################################################################
### NEW CODE TO MIGRATE TOOLS TO

sub node_info {
   my ( $group, %group_hash, %nodes, $f1, $f2);
   open(NODECFG, "< /etc/warewulf/nodes.conf");
   while(<NODECFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      $_ =~ s/^\s+//g;
      $_ =~ s/\s+$//g;
      next unless ( $_ =~ /\w+/ );
      if ( $_ =~ /^\s*\[(.+)\]\s*$/ ) {
         if ( $nodes{$1} ) {
            $node_custom = $1;
            $group = ();
         } else {
            $node_custom = ();
            $group = $1;
            $group =~ s/\s/_/g;
            %group_hash = ();
            $group_hash{"group name"} = $group;
          }
      } elsif ( $_ =~ /^\s*(node\d+)\s*=\s*(\w\w:\w\w:\w\w:\w\w:\w\w:\w\w)$/ ) {
         %{$nodes{$1}} = %group_hash;
         #%{$nodes{$1}{"group data"}} = %group_hash;
         $nodes{$1}{"nodename"} = $1;
         $nodes{$1}{"designation"} = $1;
         $nodes{$1}{"mac address"} = $2;
      } elsif ( $_ =~ /^.+=.+$/ and defined $group ) {
         ($f1, $f2) = split(/\s*=\s*/, $_);
         $group_hash{$f1} = $f2;
      } elsif ( $_ =~ /^.+=.+$/ and defined $node_custom ) {
         ($f1, $f2) = split(/\s*=\s*/, $_);
         $nodes{$node_custom}{$f1} = $f2;
         if ( $f1 eq "nodename" ) {
            $nodes{$f2} = \%{$nodes{$node_custom}};
         }
      }
   }
   return(%nodes);
}

sub group_info {
   my ( $group, %group_hash, %nodes, $f1, $f2);
   open(NODECFG, "< /etc/warewulf/nodes.conf");
   while(<NODECFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      $_ =~ s/^\s+//g;
      $_ =~ s/\s+$//g;
      next unless ( $_ =~ /\w+/ );
      if ( $_ =~ /^\s*\[(.+)\]\s*$/ ) {
         if ( $nodes{$1} ) {
            $node_custom = $1;
         } else {
            $node_custom = ();
            $group = $1;
            $group =~ s/\s/_/g;
            $group_hash{$group} = ();
          }
      } elsif ( $_ =~ /^\s*(node\d+)\s*=\s*(\w\w:\w\w:\w\w:\w\w:\w\w:\w\w)$/ ) {
         %{$nodes{$1}} = %group_hash;
         #%{$nodes{$1}{"group data"}} = %group_hash;
         $nodes{$1}{"mac address"} = $2;
      } elsif ( $_ =~ /^.+=.+$/ and ! $node_custom ) {
         ($f1, $f2) = split(/\s*=\s*/, $_);
         $group_hash{$group}{$f1} = $f2;
      }
   }
   return(%group_hash);
}

sub get_net_conf {
   my ( $device, $ifcfg_path ) = @_;
   my ( $ipaddr, $network, $netmask, $entry, $value);
   if ( ( $device =~ /^.+:\d+$/ and -f "${ifcfg_path}$device" ) or -f "${ifcfg_path}$device" ) {
      open(IFCFG,"${ifcfg_path}$device")
         or die "Could not open: ${ifcfg_path}$device\n   as defined in /etc/warewulf/master.conf!\n";
      while(<IFCFG>) {
         chomp;
         next unless $_;
         ($entry, $value) = split (/=/, $_);
         if ( $entry eq 'IPADDR' ) {
            $ipaddr = $value;
         } elsif ( $entry eq 'NETWORK' ) {
            $network = $value;
         } elsif ( $entry eq 'NETMASK' ) {
            $netmask = $value;
         }
      }
      if ( ! $ipaddr ) {
         warn "$device has no configuration for IPADDR!\n";
      }
      if ( ! $netmask ) {
         warn "$device has no configuration for NETMASK!\n";
      }
      if ( ! $network ) {
         unless ( $network = &ipcalc_network($ipaddr, $netmask) ) {
            warn "$device could not obtain the NETWORK!\n";
         }
      }
   }
   return($ipaddr, $network, $netmask);
}

sub master_info {
   my ( @config, %master, $current, $entry, $value );
   open(MASTERCFG, "< /etc/warewulf/master.conf");
   while(<MASTERCFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      $_ =~ s/^\s+//g;
      $_ =~ s/\s+$//g;
      next unless $_;
      if ( $_ =~ /^\s*\[([\w+-\.]+)\]\s*$/ ) {
         $current = $1;
      } else {
         ($entry, $value) = split(/\s+=\s+/, $_);
         $value =~ s/\s+$//g;
         $value =~ s/`(.+)`/`echo -n \`$1\``/eg;
         $master{$current}{"$entry"} = $value;
      }
   }

   if ( $master{'network'}{'admin device'} ) {
      ( $ipaddr, $network, $netmask ) = &get_net_conf(
            $master{'network'}{'admin device'}, 
            $master{'paths'}{'ifcfg'});
      $master{'ipaddrs'}{'admin'} = $ipaddr;
      $master{'network'}{'admin network'} = $network;
      $master{'network'}{'admin netmask'} = $netmask;
   } else {
      warn "You need to configure the master node's admin device!\n";
      warn "\t(/etc/warewulf/master.conf:admin device = ???)\n";
      exit 1;
   }
   if ( $master{'network'}{'cluster device'} ) {
      ( $ipaddr, $network, $netmask ) = &get_net_conf(
            $master{'network'}{'cluster device'}, 
            $master{'paths'}{'ifcfg'});
      $master{'ipaddrs'}{'cluster'} = $ipaddr;
      $master{'network'}{'cluster network'} = $network;
      $master{'network'}{'cluster netmask'} = $netmask;
   } else {
      warn "You need to configure the master node's cluster device!\n";
      warn "\t(/etc/warewulf/master.conf:cluster device = ???)\n";
      exit 1;
   }
   if ( $master{'network'}{'sharedfs device'} ) {
      ( $ipaddr, $network, $netmask ) = &get_net_conf(
            $master{'network'}{'sharedfs device'}, 
            $master{'paths'}{'ifcfg'});
      $master{'ipaddrs'}{'sharedfs'} = $ipaddr;
      $master{'network'}{'sharedfs network'} = $network;
      $master{'network'}{'sharedfs netmask'} = $netmask;
   } else {
      warn "You need to configure the master node's sharedfs device!\n";
      warn "\t(/etc/warewulf/master.conf:sharedfs device = ???)\n";
      exit 1;
   }

   if ( $master{'network'}{'boot device'} and $master{'network'}{'boot device'} ne "auto" ) {
      ( $ipaddr, $network, $netmask ) = &get_net_conf(
            $master{'network'}{'boot device'}, 
            $master{'paths'}{'ifcfg'});
   } else {
      ( $ipaddr, $network, $netmask ) = &get_net_conf(
            "$master{'network'}{'admin device'}:1", 
            $master{'paths'}{'ifcfg'});
   }
   $master{'ipaddrs'}{'boot'} = $ipaddr;
   $master{'network'}{'boot network'} = $network;
   $master{'network'}{'boot netmask'} = $netmask;

   $master{'conjoined net'}{'checksum'} = 0;
   $master{'conjoined net'}{'cluster'} = 0;
   $master{'conjoined net'}{'sharedfs'} = 0;
   if ($master{'network'}{'cluster device'} eq $master{'network'}{'admin device'}) {
      $master{'conjoined net'}{'checksum'} += 1;
      $master{'conjoined net'}{'cluster'} += 1;
   }
   if ($master{'network'}{'sharedfs device'} eq $master{'network'}{'admin device'}) {
      $master{'conjoined net'}{'checksum'} += 2;
      $master{'conjoined net'}{'sharedfs'} += 2;
   }
   if ($master{'network'}{'sharedfs device'} eq $master{'network'}{'cluster device'}) {
      $master{'conjoined net'}{'checksum'} += 4;
      $master{'conjoined net'}{'sharedfs'} += 4;
   }
   return(%master);
}

sub node_addr {
   my ($node_num, $network, $mask, @null) = @_;
   my $net_bin = unpack("N", inet_aton("$network"));
   my $mask_bin = unpack("N", inet_aton("$mask"));
   my $twiddle = ((~$mask_bin + 1) & 0xffffffff) >> 2;
   my $addr = ( $net_bin & $mask_bin ) | ( $node_num & ~$mask_bin ) | $twiddle;
   $addr = inet_ntoa(pack('N',"$addr"));
   return $addr;
}

sub alt_node_addr {
   my ($node_num, $network, $mask, @null) = @_;
   my $net_bin = unpack("N", inet_aton("$network"));
   my $mask_bin = unpack("N", inet_aton("$mask"));
   my $twiddle = ((~$mask_bin + 1) & 0xffffffff) >> 1;
   my $addr = ( $net_bin & $mask_bin ) | ( $node_num & ~$mask_bin ) | $twiddle;
   $addr = inet_ntoa(pack('N',"$addr"));
   return $addr;
}

sub ipcalc_network {
   my ( $ip, $mask, @null) = @_;
   my ( $i );
   my @ip = split /\./, $ip;
   my @mask = split /\./, $mask;
   for ($i=0; $i<4; ++$i) {
      $ip[$i] = ($ip[$i]+0) & ($mask[$i]+0);
   }
   $ip = join('.', @ip);
   return($ip);
}

sub gen_node_ips {
   my ( $ref_nodes, $ref_master) = @_;
   my %nodes = %$ref_nodes;
   my %master = %$ref_master;
   my $nodeid = -1;
   my %node_ips;
   nodeloop: foreach $node ( sort keys %nodes ) {
      next nodeloop if ( $nodes{$node}{"designation"} ne $node );
      $nodeid++;
      next nodeloop if ( $nodes{$node}{"skip"} );
      if ( $nodes{$node}{"boot ip"} ) {
         $node_ips{$node}{boot} = $nodes{$node}{"boot ip"};
      } elsif ( $nodes{$node}{"boot addr"} ){
         ($addr, $netmask) = split (/:/, $nodes{$node}{"boot addr"});
         $node_ips{$node}{boot} = $addr;
      } else {
         $node_ips{$node}{boot} = node_addr(
               $nodeid,
               $master{network}{"boot network"},
               $master{network}{"boot netmask"});
      }
# This boot2 stuff is to deal with custom MAC addresses that
# are assigned to the boot NIC that may, or may not, be there
# right at boot time.  Some NICs remember their custom MAC address
# across a warm boot.  The tulip based Linksys LNE100TX does for example.
# This simply makes the dhcpd.conf file have two different MAC addresses
# for the same node, but we have to assign an alternative IP address so
# the dhcpd won't complain...  This is needed for FNN networks. - Tim
      if ( $nodes{$node}{"boot2 ip"} ) {
         $node_ips{$node}{boot2} = $nodes{$node}{"boot2 ip"};
      } elsif ( $nodes{$node}{"boot2 addr"} ){
         ($addr, $netmask) = split (/:/, $nodes{$node}{"boot2 addr"});
         $node_ips{$node}{boot2} = $addr;
      } else {
         $node_ips{$node}{boot2} = alt_node_addr(
               $nodeid,
               $master{network}{"boot network"},
               $master{network}{"boot netmask"});
      }
      if ( $nodes{$node}{"admin ip"} ) {
         $node_ips{$node}{admin} = $nodes{$node}{"admin ip"};
      } elsif ( $nodes{$node}{"admin addr"} ){
         ($addr, $netmask) = split (/:/, $nodes{$node}{"admin addr"});
         $node_ips{$node}{admin} = $addr;
      } else {
         $node_ips{$node}{admin} = node_addr(
               $nodeid, 
               $master{network}{"admin network"}, 
               $master{network}{"admin netmask"});
      }
      if ( $nodes{$node}{"cluster ip"} ) {
         $node_ips{$node}{cluster} = $nodes{$node}{"cluster ip"};
      } elsif ( $nodes{$node}{"cluster addr"} ){
         ($addr, $netmask) = split (/:/, $nodes{$node}{"cluster addr"});
         $node_ips{$node}{cluster} = $addr;
      } else {
         $node_ips{$node}{cluster} = node_addr(
               $nodeid, 
               $master{network}{"cluster network"}, 
               $master{network}{"cluster netmask"});
      }
      if ( $nodes{$node}{"sharedfs ip"} ) {
         $node_ips{$node}{sharedfs} = $nodes{$node}{"sharedfs ip"};
      } elsif ( $nodes{$node}{"sharedfs addr"} ){
         ($addr, $netmask) = split (/:/, $nodes{$node}{"sharedfs addr"});
         $node_ips{$node}{sharedfs} = $addr;
      } else {
         $node_ips{$node}{sharedfs} = node_addr(
               $nodeid, 
               $master{network}{"sharedfs network"}, 
               $master{network}{"sharedfs netmask"});
      }
      if ( $debug ) {
         print "Generating IPs for $node: (boot=$node_ips{$node}{boot}, admin=$node_ips{$node}{admin}, cluster=$node_ips{$node}{cluster}, sharedfs=$node_ips{$node}{sharedfs}\n";
      }
   }
   return %node_ips;
}

sub gen_hosts {
   my ( $ref_nodes, $ref_master, $ref_node_ips, @null ) = @_;
   my %nodes = %$ref_nodes;
   my %master = %$ref_master;
   my %node_ips = %$ref_node_ips;
   my %used; 
   my $hsts;

   sub gen_hosts_line {
      my ( $node, $whichnet, $first, $last, $ref_ips, @null ) = @_;
      my %ips = %$ref_ips;
      my $rtn = "$ips{$whichnet}\t$first";
      if ( $ips{'admin'} eq $ips{$whichnet} ) {
         $rtn .= "${node}-admin ";
         $used{'admin'} = 1;
      }
      if ( $ips{'cluster'} eq $ips{$whichnet} ) {
         $rtn .= "${node}-clust ";
         $used{'cluster'} = 1;
      }
      if ( $ips{'sharedfs'} eq $ips{$whichnet} ) {
         $rtn .= "${node}-sfs ";
         $used{'sharedfs'} = 1;
      }
      $rtn .= "$last\n";
      return $rtn;
   }

   $hsts .= "# Do not edit below this line, it will be overwritten by Warewulf! #\n";

   $hsts .= "# Master node configuration\n";

   my $n = $master{'network'}{'nodename'};
   $used{'admin'} = 0;
   $used{'cluster'} = 0;
   $used{'sharedfs'} = 0;
   $hsts .= gen_hosts_line($n, $master{'network'}{'default network'}, "$n ", "", \%{$master{'ipaddrs'}});
   $hsts .= gen_hosts_line($n, 'admin', "", "", \%{$master{'ipaddrs'}}) unless $used{'admin'};
   $hsts .= gen_hosts_line($n, 'cluster', "", "", \%{$master{'ipaddrs'}}) unless $used{'cluster'};
   $hsts .= gen_hosts_line($n, 'sharedfs', "", "", \%{$master{'ipaddrs'}}) unless $used{'sharedfs'};
                                                                                
   $hsts .= "\n# Peer node configurations\n";

   foreach ( sort keys %node_ips ) {
      if ( $nodes{$_}{"nodename"} ) {
         $nodename = $nodes{$_}{"nodename"};
      } else {
         $nodename = $_;
      }
      $used{'admin'} = 0;
      $used{'cluster'} = 0;
      $used{'sharedfs'} = 0;
      $hsts .= gen_hosts_line($nodename, $master{'network'}{'default network'}, "$nodename ", "", \%{$node_ips{$_}});
      $hsts .= gen_hosts_line($nodename, 'admin', "", "", \%{$node_ips{$_}}) unless $used{'admin'};
      $hsts .= gen_hosts_line($nodename, 'cluster', "", "", \%{$node_ips{$_}}) unless $used{'cluster'};
      $hsts .= gen_hosts_line($nodename, 'sharedfs', "", "", \%{$node_ips{$_}}) unless $used{'sharedfs'};
   }

   return $hsts;
}


sub node_status {
   my ( $host, $port, $timeout, @NULL ) = @_;
   my ( %nodes, %nodecfg, @s, $entry, $value, $nodename, $sock );

   unless ( $host ) {
      $host = 'localhost';
   }
   unless ( $port ) {
      $port = '9873';
   }
   unless ( $timeout ) {
      $timeout = '5';
   }

   my $sock = new IO::Socket::INET ( PeerAddr => $host,
                                     PeerPort => $port,
                                     Proto    => 'tcp',
                                     Timeout  => $timeout,
                                   );
   unless ( $sock ) {
      warn "Could not connect to $host:$port!\n";
   }
   
   foreach ( keys %nodecfg ) {
      $nodes{$_}{NODESTATUS} = 'DOWN';
      $nodes{$_}{LASTCONTACT} = 'never';
   }
   
   while (<$sock>) {
      chomp;
      ( $entry, $value ) = split (/=/, $_ );
      next unless ( defined $entry and defined $value );
      if ( $entry eq 'NODE' ) {
         $nodename = $value;
         next;
      }
      if ( $nodename ) {
         $nodes{$nodename}{$entry} = $value;
      }
   }
   return(%nodes);
}


sub pxe_update {
   my ( $verbose, @null ) = @_;
   my %nodes = &node_info();
   my %master = &master_info();
   my $pxecfgdir;
   my %node_ips = &gen_node_ips(\%nodes, \%master);
   my $pxelinux_found;
   if ( $master{"paths"}{"tftpboot"} ) {
      $pxecfgdir = "$master{'paths'}{'tftpboot'}/pxelinux.cfg";
   } else {
      $pxecfgdir = "/tftpboot/pxelinux.cfg";
   }
   my $gw = "";
   my ( $boot_ip, $boot_network, $boot_netmask);
   $boot_ip = $master{'ipaddrs'}{'boot'};
   $boot_network = $master{'network'}{'boot network'};
   $boot_netmask = $master{'network'}{'boot netmask'};

   if ( -d $pxecfgdir ) {
      system("rm -f $pxecfgdir/*");
   } else {
      system("mkdir -p $pxecfgdir");
   }

   if ( ! -f "/tftpboot/pxelinux.0" ) {
      if ( $verbose ) {
         print "   Looking for pxelinux.0...\n";
      }
      open(PXELINUX,"locate pxelinux.0 2>/dev/null |");
      while(<PXELINUX>) {
         chomp;
         if ( $_ =~ /syslinux\/pxelinux\.0$/ ) {
            if ( $verbose ) {
               print "   Found $_\n";
            }
            $pxelinux_found = "1";
            system("cp -ap $_ /tftpboot/");
         }
      }
      if ( ! $pxelinux_found ) {
         if ( -f "/usr/lib/syslinux/pxelinux.0" ) {
            if ( $verbose ) {
               print "   Found /usr/lib/syslinux/pxelinux.0\n";
            }
            $pxelinux_found = "1";
            system("cp -ap /usr/lib/syslinux/pxelinux.0 /tftpboot/");
         }
      }
      if ( ! $pxelinux_found ) {
         if ( -f "/usr/lib64/syslinux/pxelinux.0" ) {
            if ( $verbose ) {
               print "   Found /usr/lib64/syslinux/pxelinux.0\n";
            }
            $pxelinux_found = "1";
            system("cp -ap /usr/lib64/syslinux/pxelinux.0 /tftpboot/");
         }
      }
      if ( ! $pxelinux_found ) {
         warn "WARNING: Could not find pxelinux.0! Please locate this, and copy\n";
         warn "         it to /tftpboot/!\n";
      }
   }

   my $pxe_template = "
   default ww
   label ww
   kernel /#kernel image#
   append rw initrd=#wwinitrd image# root=/dev/root ramdisk=15360 #warewulf#";

   if ( $verbose ) {
      print "   Adding default entry for new nodes\n";
   }

   open(PXEDEFAULT, "> $pxecfgdir/default");
   my $tmp = $pxe_template;
   $tmp =~ s/#kernel image#/kernel/g;
   $tmp =~ s/#wwinitrd image#/wwinitrd.img/g;
   $tmp =~ s/#warewulf#/warewulf=undef,$boot_ip/g;
   print PXEDEFAULT $tmp;
   close PXEDEFAULT;

   nodeloop: foreach $node ( sort keys %nodes ) {
      next nodeloop unless ( $nodes{$node}{"mac address"} =~ /\w\w:\w\w:\w\w:\w\w:\w\w/ );
      next nodeloop if ( $node ne $nodes{$node}{"designation"} );
      $vnfs = $nodes{$node}{'vnfs'};
      if ( $nodes{$node}{"nodename"} ) {
         $wwnet = "warewulf=$nodes{$node}{'nodename'},";
      } else {
         $wwnet = "warewulf=$node,";
      }
      $wwnet .= "$master{'ipaddrs'}{'admin'},";
      if ( $nodes{$node}{'gateway ip'} ) {
         $wwnet .= "$nodes{$node}{'gateway ip'},";
      } else {
         $wwnet .= "$master{'ipaddrs'}{'admin'},";
      }
      $wwnet .= "$vnfs,";
      $wwnet .= "$nodes{$node}{'admin device'}=";
      if ( $nodes{$node}{"admin addr"} ) {
         $wwnet .= $nodes{$node}{"admin addr"};
      } else {
         $wwnet .= "$node_ips{$node}{'admin'}:";
         $wwnet .= "$master{'network'}{'admin netmask'}";
      }
      if ( $nodes{$node}{'cluster device'} ne
           $nodes{$node}{'admin device'} ) {
         $wwnet .= ",";
         $wwnet .= "$nodes{$node}{'cluster device'}=";
         if ( $nodes{$node}{"cluster addr"} ) {
            $wwnet .= $nodes{$node}{"cluster addr"};
         } else {
            $wwnet .= "$node_ips{$node}{'cluster'}:";
            $wwnet .= "$master{'network'}{'cluster netmask'}";
         }
      }
      if ( $nodes{$node}{'sharedfs device'} ne
           $nodes{$node}{'admin device'} and
           $nodes{$node}{'sharedfs device'} ne
           $nodes{$node}{'cluster device'} ) {
         $wwnet .= ",";
         $wwnet .= "$nodes{$node}{'sharedfs device'}=";
         if ( $nodes{$node}{"sharedfs addr"} ) {
            $wwnet .= $nodes{$node}{"sharedfs addr"};
         } else {
            $wwnet .= "$node_ips{$node}{'sharedfs'}:";
            $wwnet .= "$master{'network'}{'sharedfs netmask'}";
         }
      }
      if ( $nodes{$node}{'other dev'} and
           $nodes{$node}{'other addr'} ) {
         $wwnet .= ",$nodes{$node}{'other dev'}=";
         $wwnet .= "$nodes{$node}{'other addr'}";
      }

   # Convert the MAC address to lower case, replace colons with dashes and
   # prepend with 01- to indicate Ethernet... all this for PXELINUX >= 2.05
      $pxeid = lc("01-$nodes{$node}{'mac address'}");
      $pxeid =~ s/:/-/g;
      if ( $verbose ) {
         print "   [$nodes{$node}{nodename}] $pxeid\n";
      }
      $tmp = $pxe_template;
      $tmp =~ s/#kernel image#/$nodes{$node}{'kernel image'}/g;
      $tmp =~ s/#wwinitrd image#/$nodes{$node}{'wwinitrd image'}/g;
      $tmp =~ s/#warewulf#/$wwnet/g;
      open(PXECFG, "> $pxecfgdir/$pxeid");
      print PXECFG $tmp;
      close PXECFG;

      if ( $nodes{$node}{'hardware addr2'} ) {
         if ( $verbose ) {
            print "   [$node] $pxeid\n";
         }
         $pxeid = lc("01-$nodes{$node}{'hardware addr2'}");
         $pxeid =~ s/:/-/g;
         open(PXECFG, "> $pxecfgdir/$pxeid");
         print PXECFG $tmp;
         close PXECFG;
      }
   }
}


sub generate_random_string {
    my $size = shift;
    my @alphanumeric = ('a'..'z', 'A'..'Z', 0..9);
    my $randstring = join '', map $alphanumeric[rand @alphanumeric], 0..$size;
    return $randstring;
}


sub write_hosts {

   my %nodes = &node_info();
   my %master = &master_info();
   my %node_ips = &gen_node_ips(\%nodes, \%master);
   my $outfile;

   open(OUTFILE, "< /etc/hosts")
      or die "Could not open /etc/hosts for reading!\n";
   while (<OUTFILE>) {
      chomp;
      if ( $_ eq '# Warewulf nodes #' ) {
         last;
      } else {
         $outfile .= "$_\n";
      }
   }
   close OUTFILE;
   
   $outfile .= "# Warewulf nodes #\n";
   $outfile .= &gen_hosts(\%nodes, \%master, \%node_ips);
   
   open(OUTFILE, "> /etc/hosts")
      or die "Could not open /etc/hosts for writing!\n";
   print OUTFILE "$outfile\n";
   close OUTFILE;
   
   my $masterIP = "$master{'ipaddrs'}{$master{'network'}{'default network'}}";
   my $null;
   my @master_addrs;
   my $iaddr;
   my $straddr;
   my $masterIP_OK;
   ( $null, $null, $null, $null, @master_addrs) = gethostbyname($master{'network'}{'nodename'});
   foreach $iaddr ( @master_addrs ) {
      $straddr = inet_ntoa($iaddr);
      if ( $straddr eq $masterIP ) {
         $masterIP_OK = 1;
      }
   }
   if ( ! $masterIP_OK ) {
      print "\n\nWarning:\n";
      print "The master's nodename '$master{'network'}{'nodename'}' resolved to $straddr.\n";
      print "It should resolve to $masterIP instead.  This should be fixed.\n";
      die "Remove the extranious entry above '# Warewlf nodes #' in /etc/hosts.\n";
   }
}







### Version 2.2.x functions...

sub node_conf_hash {
   my ( @config, %nodes, %defaults, $entry, $value );
   my $ip1 = 1;
   my $ip2 = 0;
   my $count = 0;
   open(NODECFG, "< /etc/warewulf/nodes.conf");
   while(<NODECFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      $_ =~ s/^\s+//g;
      next unless ( $_ =~ /\w+/ );
      ($entry, $value) = split(/\s+=\s+/, $_);
      $value =~ s/\s+$//g;
      $value =~ s/`(.+)`/`echo -n \`$1\``/eg;
      if ( $_ =~ /^\s*\[defaults\]\s*$/ ) {
         $node = ();
         $default = '1';
      } elsif ( $_ =~ /^\s*\[([\w+-\.]+)\]\s*$/ ) {
         $node = $1;
         $default = ();
         %{$nodes{$node}} = %defaults;
      } elsif ( $node ) {
         $nodes{$node}{"$entry"} = $value;
      } elsif ( $default ) {
         $defaults{"$entry"} = $value;
      }
   }
   foreach $node ( sort keys %nodes ) {
      $nodes{$node}{'conjoined net'}{'checksum'} = 0;
      $nodes{$node}{'conjoined net'}{'cluster'} = 0;
      $nodes{$node}{'conjoined net'}{'sharedfs'} = 0;
      if ($nodes{$node}{'cluster device'} eq $nodes{$node}{'admin device'}) {
         $nodes{$node}{'conjoined net'}{'checksum'} += 1;
         $nodes{$node}{'conjoined net'}{'cluster'} += 1;
         if ( ! $nodes{$node}{'admin ip'} ) {
            $nodes{$node}{'admin ip'} = $nodes{$node}{'cluster ip'};
         }
      }
      if ($nodes{$node}{'sharedfs device'} eq $nodes{$node}{'admin device'}) {
         $nodes{$node}{'conjoined net'}{'checksum'} += 2;
         $nodes{$node}{'conjoined net'}{'sharedfs'} += 2;
         if ( ! $nodes{$node}{'admin ip'} ) {
            $nodes{$node}{'admin ip'} = $nodes{$node}{'sharedfs ip'};
         }
      }
      if ($nodes{$node}{'sharedfs device'} eq $nodes{$node}{'cluster device'}) {
         $nodes{$node}{'conjoined net'}{'checksum'} += 4;
         $nodes{$node}{'conjoined net'}{'sharedfs'} += 4;
         if ( ! $nodes{$node}{'cluster ip'} ) {
            $nodes{$node}{'cluster ip'} = $nodes{$node}{'sharedfs ip'};
         }
      }
   }
   return(%nodes);
}

sub vnfs_conf_hash {
   my ( @config, %vnfss, %defaults, $vnfs, $entry, $value, $default );
   open(VNFSCFG, "< /etc/warewulf/vnfs.conf");
   while(<VNFSCFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      $_ =~ s/^\s+//g;
      next unless $_;
      ($entry, $value) = split(/\s+=\s+/, $_);
      $value =~ s/\s+$//g;
      $value =~ s/`(.+)`/`echo -n \`$1\``/eg;
      if ( $_ =~ /^\s*\[defaults\]\s*$/ ) {
         $vnfs = ();
         $default = '1';
      } elsif ( $_ =~ /^\s*\[([\w+-\.]+)\]\s*$/ ) {
         $vnfs = $1;
         $default = ();
         %{$vnfss{$vnfs}} = %defaults;
      } elsif ( $vnfs ) {
         $vnfss{$vnfs}{"$entry"} = $value;
      } elsif ( $default ) {
         $defaults{"$entry"} = $value;
      }
   }
   return(%vnfss);
}




sub node_conf_sanity {
   my %nodes = @_;
   my @required = (
      "hardware addr",
      "boot device",
      "admin device",
      "sharedfs device",
      "cluster device",
      "vnfs",
   );
   nodeloop: foreach $node ( sort keys %nodes ) {
      next nodeloop if ( $nodes{$node}{'skip'} );
      foreach $req ( @required ) {
         unless ( $nodes{$node}{"$req"} ) {
            warn "NODE ERROR: $node is missing required configuration: $req!\n";
         }
      }
   }
}

sub vnfs_conf_sanity {
   my %vnfss = @_;
   my @required = (
      "path",
      "kernel",
      "kernel version",
      "kernel modules",
      "exclude",
      "master includes",
   );
   foreach $vnfs ( sort keys %vnfss ) {
      foreach $req ( @required ) {
         unless ( $vnfss{$vnfs}{"$req"} ) {
            warn "VNFS ERROR: $vnfs is missing required configuration: $req!\n";
         }
      }
   }
}





sub def_vnfs {
   my ( %vnfs ) = @_;
   foreach ( keys %vnfs ) {
      if ( $vnfs{$_}{'default'} == '1' ) {
         return $_;
      }
      # Make the first one the default if none are specified as default
      $first = $_ unless $first;
   }
   return $first;
}


sub users_filter {
# PARAMS: None
# RETURN: Hash of nodes that the user does not want (filtered out)
# DESC:   Looks for the users enviornment variable for '$NODES' and then checks 
#         to see if it is a node list (\W delim) or a file of defined nodes 
#         that the user wants to view.

   my ( %nodelist);
   if ( $ENV{NODES} ) {
      if ( -f $ENV{NODES} ) {
         open(NODES, "< $ENV{NODES}");
         while (<NODES>) {
            chomp;
            $_ =~ s/#.*$//;
            $_ =~ s/\W.*$//;
            next unless $_;
            $nodelist{$_} = '1';
         }
         close $NODES;
      } else {
         @s = split(/\W/, $ENV{NODES});
         foreach ( @s ) {
            $nodelist{$_} = '1';
         }
      }
   }
   return(%nodelist);
}

sub nodes_access {
# PARAMS: Array returned from the function &read_nodecfg
# RETURN: Hash that contains boolean true values for each node that the
#         current user has access to.

   my %nodecfg = @_;
   my ( %group );
   my ( $users );
   
   while ( @line = getgrent() ) {
      $users = $line[3];
      $users =~ s/\s/,/g;
      $group{$line[0]} = $users;
   }
   foreach ( keys %nodecfg ) {
      @u = split(/\s*,\s*/, $nodecfg{$_}{users});
      foreach $user ( @u ) {
         if ( $user eq getpwuid($<) ) {
            $nodes{$_} = '1';
         }
      }
      @g = split(/\s*,\s*/, $nodecfg{$_}{groups});
      foreach $grp ( @g ) {
         @u = split(/\s*,\s*/, $group{$grp});
         foreach $user ( @u ) {
            if ( $user eq getpwuid($<) ) {
               $nodes{$_} = '1';
            }
         }
      }
   }
   return(%nodes);
}

sub client_conf {
   my %configs, $entry, $value;
   open(TOOLSCONF, "< /etc/warewulf/client.conf")
      or warn "Could not open /etc/warewulf/client.conf!\n";
   while (<TOOLSCONF>) {
      chomp;
      $_ =~ s/#.*$//;
      $_ =~ s/^\s+//g;
      next unless $_;
      ($entry, $value) = split(/\s+=\s+/, $_);
      $configs{$entry} = "$value";
   }
   close $NODES;
   if ( $ENV{WAREWULF_MASTER} ) {
      $configs{'warewulf master'} = $ENV{WAREWULF_MASTER};
   }
   return(%configs);
}



##############################################################################
### PRE WW\2.0 functions

sub map_myri {
# PARAMS: None
# RETURN: Exit value of the myrinet mapper
# DESC:   Needed to map the myrinet network

   my ( $verbose, @NULL ) = @_;
   if ( $verbose ) {
      print "Mapping the Myrinet network\n";
   }
   system("$config::myrinet_mapper $config::myrinet_mapper_args 2>&1 > /var/log/myrinet_mapper.log");
   return($?);
}


sub read_nodecfg {
# PARAMS: None:
# RETURN: Array containing the node configuration file.
# DESC:   Nothing complex here... Just returning an array of the config file

   my ( @config );
   open(NODECFG, "< /etc/warewulf/node.conf");
   while(<NODECFG>) {
      chomp;
      push(@config, $_);
   }
   return(@config);
}

sub read_mastercfg {
# PARAMS: None:
# RETURN: Array containing the master configuration file.
# DESC:   Nothing complex here... Just returning an array of the config file

   my ( @config );
   open(MASTERCFG, "< /etc/warewulf/master.conf");
   while(<MASTERCFG>) {
      chomp;
      push(@config, $_);
   }
   return(@config);
}

sub node_config {
# PARAMS: Array returned from the function &read_nodecfg
# RETURN: Hashed hash containing the configuration of the slave nodes in 
#         the form $nodecfg{nodename}{config}.

   my @config = @_;
   my ( %nodecfg, $node, @s, $entry, $value );
   foreach ( @config ) {
      $_ =~ s/#.*$//g;
      next unless ( $_ =~ /\w/ );
      if ( $_ =~ /^node.+$/ ) {
         @s = split(/\s+/, $_);
         $node = @s[1];
      } else {
         $_ =~ s/^\s+//g;
         ( $entry, $value ) = split(/\s*=\s*/, $_);
         $nodecfg{$node}{$entry} = $value;
      }
   }
   return (%nodecfg);
}

sub master_config {
# PARAMS: Array returned from the function &read_mastercfg
# RETURN: Hashed hash containing the configuration of the master nodes in
#         the form $mastercfg{masterjob}{config} where 'masterjob' is type of
#         master node (ie. primary_master, dhcp_master, etc...).
   
   my @config = @_;
   my ( %mastercfg, $master, @s, $entry, $value );
   foreach ( @config ) {
      chomp;
      $_ =~ s/#.*$//g;
      next unless ( $_ =~ /\w/ );
      if ( $_ =~ /^master.+$/ ) {
         @s = split(/\s+/, $_);
         $master = @s[1];
      } else {
         $_ =~ s/^\s+//g;
         ( $entry, $value ) = split(/\s*=\s*/, $_);
         $mastercfg{$master}{$entry} = $value;
      }
   }
   return(%mastercfg);
}

sub last_node {
# PARAMS: Array returned from the function &read_nodecfg
# RETURN: Scaler value of the next node name (incremented the highest numbered
#         node by one.
# DESC:   This function is used when configuring a new node.

   my @config = @_;
   foreach ( reverse @config ) {
      $_ =~ s/#.*$//g;
      next unless ( $_ =~ /\w/ );
      if ( $_ =~ /^node.+$/ ) {
         @s = split(/\s+/, $_);
         $node = @s[1];
         last;
      }
   }
   return($node);
}

sub primary_master {
# PARAMS: Array returned from the function &read_mastercfg
# RETURN: Scaler value of the primary master node
# DESC:   Find the primary master node in the list of masters

   my %mastercfg = &master_config(@_);
   foreach ( keys %mastercfg ) {
      return($_) if ( $mastercfg{$_}{primary} eq 'yes' );
   }
}

sub tftp_master {
# PARAMS: Array returned from the function &read_mastercfg
# RETURN: Scaler value of the TFTP master node
# DESC:   Find the master node that is providing the TFTP services to download
#         the node boot image.

   my %mastercfg = &master_config(@_);
   foreach ( keys %mastercfg ) {
      return($_) if ( $mastercfg{$_}{tftpdir} );
   }
}

sub dhcp_master {
# PARAMS: Array returned from the function &read_mastercfg
# RETURN: Scaler value of the dhcpd master node
# DESC:   Find the master node that is providing the DHCPD services to the 
#         slave nodes.

   my %mastercfg = &master_config(@_);
   foreach ( keys %mastercfg ) {
      return($_) if ( $mastercfg{$_}{dhcpcfg} );
   }
}

sub nfs_master {
# PARAMS: Array returned from the function &read_mastercfg
# RETURN: Scaler value of the primary master node
# DESC:   Find the master node that is providing the NFS export.

   my %mastercfg = &master_config(@_);
   foreach ( keys %mastercfg ) {
      return($_) if ( $mastercfg{$_}{nfsvol} );
   }
}

sub syslog_master {
# PARAMS: Array returned from the function &read_mastercfg
# RETURN: Scaler value of the primary master node
# DESC:   Find the master node that is configured to recieve syslogd

   my %mastercfg = &master_config(@_);
   foreach ( keys %mastercfg ) {
      return($_) if ( $mastercfg{$_}{syslogd} eq 'yes' );
   }
}

sub build_hosts {
# PARAMS: None
# RETURN: None
# DESC:   This function will rewrite /etc/hosts from the line 
#         "# Warewulf nodes #" and below with all of the current nodes.

   my $hosts;
   my %nodecfg = &node_config(@_);
   #print "Re-writing '/etc/hosts'\n";
   open(HOSTS, "< /etc/hosts")
      or die "Could not open /etc/hosts for reading!\n";
   while (<HOSTS>) {
      chomp;
      if ( $_ eq '# Warewulf nodes #' ) {
         $warewulf_nodes = '1';
         last;
      } else {
         $hosts .= "$_\n";
      }
   }
   close HOSTS;
   $hosts .= "# Warewulf nodes #\n";
   foreach ( sort keys %nodecfg ) {
      $hosts .= "$nodecfg{$_}{clust_ip}\t$_\n";
      if ( $nodecfg{$_}{admin_ip} ne $nodecfg{$_}{clust_ip} ) {
         $hosts .= "$nodecfg{$_}{admin_ip}\tadmin-$_\n";
      }
   }
   open(HOSTS, "> /etc/hosts") 
      or die "Could not open /etc/hosts for writing!\n";
   print HOSTS "$hosts\n";
   close HOSTS;
}

sub next_node {
# PARAMS: This function takes two arrays as a paramater. The only way to do 
#         this is with array references. This means you MUST backslash the
#         two arrays when you call the function!!!
#         (ie. '&next_node(\@foo,\@poo);')
   my %return;
   my ( $ref_node, $ref_master) = @_;
   my @node = @$ref_node;
   my @master = @$ref_master;
   my $last_node = &last_node(@node);
   my $primary_master = &primary_master(@master);
   my %nodecfg = &node_config(@node);
   my %mastercfg = &master_config(@master);
   foreach ( keys %nodecfg ) {
      $return{clust_ip} = $nodecfg{$_}{clust_ip}
         if ( defined &ipsize($nodecfg{$_}{clust_ip},$return{clust_ip}));
      $return{admin_ip} = $nodecfg{$_}{admin_ip}
         if ( defined &ipsize($nodecfg{$_}{admin_ip},$return{admin_ip}));
   }
   foreach ( keys %mastercfg ) {
      $return{clust_ip} = $mastercfg{$_}{clust_ip}
         if ( defined &ipsize($mastercfg{$_}{clust_ip},$return{clust_ip}));
      $return{admin_ip} = $mastercfg{$_}{admin_ip}
         if ( defined &ipsize($mastercfg{$_}{admin_ip},$return{admin_ip}));
   }
   $return{users} = $nodecfg{$last_node}{users};
   $return{groups} = $nodecfg{$last_node}{groups};
      @s = split(/\./, $return{admin_ip});
      $return{admin_ip} = "$s[0].$s[1].$s[2].". ( $s[3] + 1 );
      @s = split(/\./, $return{clust_ip});
      $return{clust_ip} = "$s[0].$s[1].$s[2].". ( $s[3] + 1 );
   if ( $last_node ) {
      $return{clust_dev} = $nodecfg{$last_node}{clust_dev};
      $w = $n = $last_node ;
      $w =~ s/\d//g;
      $n =~ s/\D//g;
      if ( $n ) {
         $n ++;
      } else {
         $n = '001';
      }
      $return{node} = $w . $n;
   } else {
      if ( $mastercfg{$primary_master}{clust_dev} eq $mastercfg{$primary_master}{admin_dev} ) {
         $return{clust_dev} = 'na';
         $return{clust_ip} = $return{admin_ip};
      } else {
         $return{clust_dev} = $mastercfg{$primary_master}{clust_dev};
      }
      $return{node} = 'node001';
   }
   return(%return);
}


###############################################################################
### CODE I AM NOT MODIFYING

sub ipsize {
   my ( $newip, $oldip ) = @_;
   my @new = split(/\./, $newip);
   my @old = split(/\./, $oldip);
   for(my $i; $i <= 3; $i++) {
      return     if ( $new[$i] < $old[$i] );
      return '1' if ( $new[$i] > $old[$i] );
   }

}


###############################################################################
### OLD CODE TO DO AWAY WITH BELOW


sub node_cfgdisabled {
my @s = ();
   open(NODECFG, "< /etc/warewulf/node.conf");
#      or warn "could not open /etc/warewulf/node.conf\n\n";
   while(<NODECFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      next unless ( $_ =~ /\w/ );
      if ( $_ =~ /^node.+$/ ) {
         @s = split(/\s+/, $_);
         $node = @s[1];
         $total_nodes ++;
         next;
      }
      $_ =~ s/^\s+//g;
      ( $entry, $value ) = split(/\s*=\s*/, $_);
      $nodecfg{$node}{$entry} = $value;
      $high_cluster_ip = $value if ( $entry eq 'clust_ip' and defined &ipsize($value,$ipaddr));
      $high_admin_ip = $value if ( $entry eq 'admin_ip' and defined &ipsize($value,$ipaddr));
      $knownmac{$value} = '1' if ( $entry eq 'mac' );
      $node_clust_dev = $value if ( $entry eq 'clust_dev' );
      $node_admin_dev = $value if ( $entry eq 'admin_dev' );
      $groups = $value if ( $entry eq 'groups' );
      $users = $value if ( $entry eq 'users' );
   }
   close NODECFG;
   open(MASTERCFG, "< /etc/warewulf/master.conf") ;
   while(<MASTERCFG>) {
      chomp;
      $_ =~ s/#.*$//g;
      next unless ( $_ =~ /\w/ );
      if ( $_ =~ /^master.+$/ ) {
         @s = split(/\s+/, $_);
         $master = @s[1];
         next;
      }
      $_ =~ s/^\s+//g;
      ( $entry, $value ) = split(/\s*=\s*/, $_);
      $mastercfg{$master}{$entry} = $value;
      #$highipaddr = $value if ( $entry eq 'ipaddr' and defined &ipsize($value,$ipaddr));
      unless ( $high_cluster_ip ) {
         $high_cluster_ip = $value if ( $entry eq 'clust_ip' and defined &ipsize($value,$ipaddr));
      }
      unless ( $high_admin_ip ) {
         $high_admin_ip = $value if ( $entry eq 'admin_ip' and defined &ipsize($value,$ipaddr));
      }
      $tftp_master = $master if ( $entry eq 'bootimg' );
      $dhcp_master = $master if ( $entry eq 'dhcpcfg' );
      $nfs_master = $master if ( $entry eq 'nfsvol' );
      $syslog_master = $master if ( $entry eq 'syslogd' and $value eq 'yes');
      $primary_master = $master if ( $entry eq 'primary' and $value eq 'yes');
   }
   $mastercfg{$tftp_master}{bootimg} = "/$mastercfg{$tftp_master}{bootimg}"
      unless ( $mastercfg{$tftp_master}{bootimg} =~ /^\// );
   close MASTERCFG;
}

#sub nodes_up {
#   open(NODESUP, "< /etc/warewulf/node.status");
#   while (<NODESUP>) {
#      chomp;
#      $_ =~ s/#.*$//;
#      next unless $_;
#      @s = split (/\s/, $_);
#      if ( $s[1] eq 'READY' ) {
#         $nodesup{$s[0]} = '1';
#         $nodeup_count ++;
#      }
#   }
#}


1;
