#!/usr/bin/perl
#
# This script is a component of Warewulf,
# http://www.runlevelzero.net/greg/warewulf
#
#########################################################################
#
# Copyright (c) 2003, The Regents of the University of California, through
# Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The GNU GPL Document can be found at:
# http://www.gnu.org/copyleft/gpl.html
#
#########################################################################
#
# Written and maintained by:
#       Greg Kurtzer, <gmkurtzer@lbl.gov>


use File::Basename;
use File::Path;
use File::Copy;
use Getopt::Long;

GetOptions(
   'print'       => \$print,
   'copy=s'      => \$copy,
   'help'        => \$help,
);

if ( $help or ( ! $print and ! $copy )) {
   print "USAGE: $0 [opts] [binaries to check]...\n\n";
   print "   --print        Print the list of library dependancies for all binaries\n";
   print "   --copy=/prefix Copy the library and links to the location\n";
   print "   --help         This summary\n\n";
   exit;
}

sub build_deps {
   my ($bin, @NULL) = @_;
   my @return;
   return if ( ! -f "$bin" );
   open(DEPS, "ldd $bin |");
   while(<DEPS>) {
      chomp;
      @f = split(" => ", $_);
      @l = split(/\s+/, $f[1]);
      if ( -f "$l[0]" ) {
         $foo = readlink("$l[0]");
         $link = basename("$l[0]");
         $file = basename("$foo");
         $dir = dirname("$l[0]");
         if ( !  $deps{"$dir/$foo"} ) {
            push(@return, "$dir/$foo" );
         }
         $deps{"$dir/$foo"}{LINK} = $link;
         $deps{"$dir/$foo"}{FILE} = $file;
         $deps{"$dir/$foo"}{PATH} = $dir;
      }
   }
   return(@return);
}


foreach $checkfile ( @ARGV ) {
   push (@ARGV, &build_deps($checkfile));
}

foreach $lib ( sort keys %deps ) {
   if ( $print ) {
      print "$deps{$lib}{PATH}/$deps{$lib}{FILE}\n";
   }
   if ( $copy ) {
      mkpath("$copy/$deps{$lib}{PATH}");
      copy("$deps{$lib}{PATH}/$deps{$lib}{FILE}", "$copy/$deps{$lib}{PATH}/$deps{$lib}{FILE}");
      system("strip $copy/$deps{$lib}{PATH}/$deps{$lib}{FILE}");
      if ( $deps{$lib}{FILE} ne $deps{$lib}{LINK} ) {
         symlink("$deps{$lib}{FILE}", "$copy/$deps{$lib}{PATH}/$deps{$lib}{LINK}");
      }
   }
}
