#!/usr/bin/perl


#
#  vnfs-remote.pl version 1.1
#########################################################################
#
# Copyright (c) 2003, The Regents of the University of California, through
# Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# The GNU GPL Document can be found at:
# http://www.gnu.org/copyleft/gpl.html
#
#########################################################################
#
#  Tue May 24 12:17:10 PDT 2005
#  galitz at berkeley.edu
#  geoff at galitz.org
#
#  This script will create a VNFS out of an installed filesystem on 
#  a system via rsync.  The system can be of any type.  The image name
#  and the rsync module are assumed to be the same.
#
#  Note:  The VNFS will not be compressed and can get quite large
#	  The /proc and /sys filesystems are not transmitted (the
#	  mountpoints are).
#
#  To do items:  
#	- Add support to override rsync-module
#	- Add support for multple rsync daemons (in case the master is 
#	  already running one).
#
#

use Getopt::Long;


GetOptions(
	'receive'	=> \$receive,
	'pull'		=> \$pull,
	'rsync-module'	=> \$rmod,
	'help'        => \$help,
);


$image = @ARGV[0];
$remote_host = @ARGV[1];
$local_image = "/vnfs/$image";

if ( $mod ) {
	print "Remote module is $mod\n";
}

# are we receiving or getting?
if ( $pull or $receive ) {
	if ( ! -d "/vnfs/$image" ) {
		print "Creating /vnfs/$image\n";
		system ("mkdir /vnfs/$image");
	}
# do we need to create a default VNFS?
	if (  ! -d "/vnfs/default" ) {
		system ("ln -s $local_image /vnfs/default");
	} 
}


# to receive, we setup our own rsync daemon and wait for 
# remote system to connect to us

if ( $receive ) {
	print "Ok...  receive\n";
	# create the receive configuration
	open ( REC_CONF, "> /tmp/rsyncd.conf.warewulf");
	print REC_CONF "[ $image ]\n";
	print REC_CONF "path=$local_image\n";
	print REC_CONF	"read only = false\n";
	print REC_CONF "uid = 0\n";
	print REC_CONF "exclude = /proc/* /sys/* \n";
	close REC_CONF;
	print "I will now launch an rsync server process.\n";
	print "When the transfer is complete, type Ctrl-C to proceed.\n";
   	print "Waiting for remote image...";
   	system ("/usr/bin/rsync --daemon --no-detach --config=/tmp/rsyncd.conf.warewulf ");
	system ("rm /tmp/rsyncd.conf.warewulf");
	print "Done!\n";
	exit;
}


# for pulling, we will call rsync and make a connection to 
# to the remote system.

if ( $pull ) {
	if ( ! "$remote_host" and "image" ) {
		print "I need a remote host and image to pull from.\n";
		print "vnfs-remote.pl --help for more info.\n";
		exit 1;
	}
	print "Ok... pull\n";
	print ("Will execute :  rsync -avz --exclude=/proc/* --exclude=/sys/* $remote_host\:\:$image $local_image\n");
	system ("cd $local_image ; rsync -avz --exclude=/proc/* --exclude=/sys/* $remote_host\:\:$image .");
	print "Done!\n";
	exit;
}


if ( $help ) {
	print "Usage: vnfs-remote.pl [options] image remote-host \n\n";
	print "Options:\n";
	print " --pull         retrieves a remote image\n";
	print " --receive      receives a remote image from a remote system\n\n";
	print "vnfs-remote will create a VNFS out of an installed filesystem for you via rsync.\n";
	print "The image name you use here will be the name of the VNFS image (in /vnfs) and\n";
	print "will be the name of rsync module to connect to.\n";
	print "\nFor example,\n";
	print "vnfs-remote --pull custom-image test.whatever.com \n";
	print "vnfs-remote --receive custom-image\n";
	print "\n\n";
	exit;
}

print "vnfs-remote --help for usage information.\n";
